#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_IPOPT
#undef HAVE_IPOPT
#endif

#include <dune/common/bitsetvector.hh>

#include <dune/fufem/assemblers/transferoperatorassembler.hh>
#include <dune/solvers/solvers/solver.hh>

#include "solverfactory_old.hh"


template <class DeformedGrid, class Nonlinearity, class MatrixType, class VectorType>
SolverFactoryOld<DeformedGrid, Nonlinearity, MatrixType, VectorType>::SolverFactoryOld(
    const Dune::ParameterTree& parset, const Dune::Contact::NBodyAssembler<DeformedGrid, VectorType>& nBodyAssembler,
    const Dune::BitSetVector<DeformedGrid::dimension>& ignoreNodes)
    : nBodyAssembler_(nBodyAssembler),
      baseEnergyNorm_(baseSolverStep_),
      baseSolver_(&baseSolverStep_,
                       parset.get<size_t>("linear.maximumIterations"),
                       parset.get<double>("linear.tolerance"), &baseEnergyNorm_,
                       Solver::QUIET),
      transferOperators_(nBodyAssembler.getGrids().at(0)->maxLevel()) {

 /* baseSolverStep_.obstacles_ = using HasObstacle = Dune::BitSetVector<BlockSize>;
    using Obstacle = BoxConstraint<Field, BlockSize>;
    const HasObstacle* hasObstacle_;
    const std::vector<Obstacle>* obstacles_;*/

  multigridStep_ = std::make_shared<Step>();
  // tnnmg iteration step
  multigridStep_->setMGType(parset.get<int>("main.multi"), parset.get<int>("main.pre"), parset.get<int>("main.post"));
  multigridStep_->setIgnore(ignoreNodes);
  multigridStep_->setBaseSolver(baseSolver_);
  multigridStep_->setSmoother(&presmoother_, &postsmoother_);
  multigridStep_->setHasObstacle(nBodyAssembler_.totalHasObstacle_);
  multigridStep_->setObstacles(nBodyAssembler_.totalObstacles_);
  multigridStep_->setVerbosity(NumProc::QUIET);

  // create the transfer operators
  const int nCouplings = nBodyAssembler_.nCouplings();
  const auto grids = nBodyAssembler_.getGrids();

  for (size_t i=0; i<transferOperators_.size(); i++) {
    transferOperators_[i] = std::make_shared<TransferOperator>();
  }

  std::vector<const Dune::BitSetVector<1>*> fineHasObstacle(nCouplings);
  std::vector<const MatrixType*> mortarTransfer(nCouplings);
  std::vector<std::array<int,2> > gridIdx(nCouplings);

  for (int j=0; j<nCouplings; j++) {
    fineHasObstacle[j]    = nBodyAssembler_.contactCoupling_[j]->nonmortarBoundary().getVertices();
    mortarTransfer[j] = &nBodyAssembler_.contactCoupling_[j]->mortarLagrangeMatrix();
    gridIdx[j]        = nBodyAssembler_.coupling_[j].gridIdx_;
  }

  TransferOperator::setupHierarchy(transferOperators_,
                                   grids,
                                   mortarTransfer,
                                   nBodyAssembler_.localCoordSystems_,
                                   fineHasObstacle,
                                   gridIdx);

  multigridStep_->setTransferOperators(transferOperators_);
}

template <class DeformedGrid, class Nonlinearity, class MatrixType, class VectorType>
auto SolverFactoryOld<DeformedGrid, Nonlinearity, MatrixType, VectorType>::getStep()
    -> std::shared_ptr<Step> {
  return multigridStep_;
}

#include "solverfactory_tmpl_old.cc"
